﻿Imports System
Imports System.Collections.Generic
Imports System.IO
Imports System.Collections

' Solves the max-clique problem using a greedy algorithm.
' A clique of a graph is a subset where every node is connected to all
' other nodes in the subset. The max-clique problem is to find the
' largest clique. Max-clique is NP-complete.

Namespace GreedyMaxClique
    Module GreedyMaxCliqueProgram
        Dim rnd As Random = Nothing

        Public Sub Main(args As String())
            Try
                Console.WriteLine(vbLf & "Begin greedy maximum clique demo" & vbLf)

                Dim graphFile As String = "DimacsGraph.clq"
                ' small demo file. best (largest) clique size is 4.
                'string graphFile = "..\\..\\C2000.9.clq"; // large DIMACS benchmark challenge file. best (largest) known clique is 76.
                Console.WriteLine("Loading graph into memory")
                Console.WriteLine("Graph loaded and validated" & vbLf)
                Dim graph As New MyGraph(graphFile, "DIMACS")

                Dim maxTime As Integer = 20
                ' use 20 for small demo file; use about 1,000-10,000 for benchmark problem.
                Dim targetCliqueSize As Integer = graph.NumberNodes
                ' max-clique size for any graph is the size of the graph -- graph is one giant clique.
                Dim maxClique As List(Of Integer) = FindMaxClique(graph, maxTime, targetCliqueSize)
                Console.WriteLine(vbLf & "Maximum time reached")
                Console.WriteLine(vbLf & "Size of best clique found = " & maxClique.Count.ToString())

                Console.WriteLine(vbLf & "Best clique found:")
                Console.WriteLine(ListAsString(maxClique))

                Console.WriteLine(vbLf & "End greedy maximum clique demo" & vbLf)
                Console.ReadLine()
            Catch ex As Exception
                Console.WriteLine("Fatal: " & ex.Message)
                Console.ReadLine()
            End Try
        End Sub ' Main

        Private Function FindMaxClique(graph As MyGraph, maxTime As Integer, targetCliqueSize As Integer) As List(Of Integer)
            ' you may want to comment-out WriteLine statements for large graphs.
            Dim clique As New List(Of Integer)()
            rnd = New Random(1)
            Dim time As Integer = 0
            Dim timeBestClique As Integer = 0
            Dim timeRestart As Integer = 0
            Dim nodeToAdd As Integer = -1
            Dim nodeToDrop As Integer = -1

            Dim rndNode As Integer = rnd.[Next](0, graph.NumberNodes)
            Console.WriteLine("Adding node " & rndNode.ToString())
            clique.Add(rndNode)

            Dim bestClique As New List(Of Integer)()
            bestClique.AddRange(clique)
            Dim bestSize As Integer = bestClique.Count
            timeBestClique = time

            Dim possibleAdd As List(Of Integer) = MakePossibleAdd(graph, clique)
            ' nodes which will increase size of clique
            Dim oneMissing As List(Of Integer) = MakeOneMissing(graph, clique)

            While time < maxTime AndAlso bestSize < targetCliqueSize
                time += 1

                Dim cliqueChanged As Boolean = False
                ' to control branching logic in loop
                If possibleAdd.Count > 0 Then
                    ' 
                    nodeToAdd = GetNodeToAdd(graph, possibleAdd)
                    ' node from possibleAdd which is most connected to nodes in possibleAdd
                    Console.WriteLine("Adding node " & nodeToAdd.ToString())
                    clique.Add(nodeToAdd)
                    clique.Sort()
                    cliqueChanged = True
                    If clique.Count > bestSize Then
                        bestSize = clique.Count
                        bestClique.Clear()
                        bestClique.AddRange(clique)
                        timeBestClique = time
                    End If
                End If
                ' add node
                If cliqueChanged = False Then
                    If clique.Count > 0 Then
                        nodeToDrop = GetNodeToDrop(graph, clique, oneMissing)
                        ' find node in clique which generate max increase in possibleAdd set
                        'if more than one, pick one at random
                        Console.WriteLine("Dropping node " & nodeToDrop.ToString())
                        clique.Remove(nodeToDrop)
                        clique.Sort()
                        cliqueChanged = True
                    End If
                End If
                ' drop node
                Dim restart As Integer = 2 * bestSize
                If time - timeBestClique > restart AndAlso time - timeRestart > restart Then
                    ' restart? if it's been a while since we found a new best clique or did a restart . . .
                    Console.WriteLine(vbLf & "Restarting" & vbLf)
                    timeRestart = time
                    Dim seedNode As Integer = rnd.[Next](0, graph.NumberNodes)
                    clique.Clear()
                    Console.WriteLine("Adding node " & seedNode.ToString())
                    clique.Add(seedNode)
                End If ' restart
                possibleAdd = MakePossibleAdd(graph, clique)

                'ValidateState(graph, clique, possibleAdd, oneMissing)

                oneMissing = MakeOneMissing(graph, clique)
            End While ' main processing loop

            Return bestClique
        End Function ' FindMaxClique
        
        Private Function MakePossibleAdd(graph As MyGraph, clique As List(Of Integer)) As List(Of Integer)
            ' create list of nodes in graph which are connected to all nodes in clique and therefore will form a larger clique
            ' calls helper FormsALargerClique
            Dim result As New List(Of Integer)()

            For i As Integer = 0 To graph.NumberNodes - 1
                ' each node in graph
                If FormsALargerClique(graph, clique, i) = True Then
                    result.Add(i)
                End If
            Next
            Return result
            ' could be empty List
        End Function ' MakePossibleAdd

        Private Function FormsALargerClique(graph As MyGraph, clique As List(Of Integer), node As Integer) As Boolean
            ' is node connected to all nodes in clique?
            For i As Integer = 0 To clique.Count - 1
                ' compare node against each member of clique
                'If (clique(i) = node) Then Return False 'node is already in clique so node will not form a larger clique
                If graph.AreAdjacent(clique(i), node) = False Then
                    Return False
                    ' node is not connected to one of the nodes in the clique
                End If
            Next
            Return True
            ' passed all checks
        End Function ' FormsALargerClique

        Private Function GetNodeToAdd(graph As MyGraph, possibleAdd As List(Of Integer)) As Integer
            ' find node from a List of allowed and possible add which has max degree in posibleAdd
            ' there could be more than one, if so, pick one at random
            'If (IsNothing(possibleAdd)) Then Throw New Exception("List possibleAdd is null in GetNodeToAdd")
            'If (possibleAdd.Count = 0) Then Throw New Exception("List possibleAdd has Count 0 in GetNodeToAdd")

            If possibleAdd.Count = 1 Then
                ' there is only 1 node to choose from
                Return possibleAdd(0)
            End If

            ' examine each node in possibleAdd to find the max degree in possibleAdd 
            '(because there could be several nodes tied with max degree)
            Dim maxDegree As Integer = 0
            For i As Integer = 0 To possibleAdd.Count - 1
                Dim currNode As Integer = possibleAdd(i)
                Dim degreeOfCurrentNode As Integer = 0
                For j As Integer = 0 To possibleAdd.Count - 1
                    ' check each node in possibleAdd
                    Dim otherNode As Integer = possibleAdd(j)
                    If graph.AreAdjacent(currNode, otherNode) = True Then
                        degreeOfCurrentNode += 1
                    End If
                Next
                If degreeOfCurrentNode > maxDegree Then
                    maxDegree = degreeOfCurrentNode
                End If
            Next

            ' now rescan possibleAdd and grab all nodes which have maxDegree
            Dim candidates As New List(Of Integer)()
            For i As Integer = 0 To possibleAdd.Count - 1
                Dim currNode As Integer = possibleAdd(i)
                Dim degreeOfCurrentNode As Integer = 0
                For j As Integer = 0 To possibleAdd.Count - 1
                    ' check each node in possibleAdd
                    Dim otherNode As Integer = possibleAdd(j)
                    If graph.AreAdjacent(currNode, otherNode) = True Then
                        degreeOfCurrentNode += 1
                    End If
                Next
                If degreeOfCurrentNode = maxDegree Then
                    candidates.Add(currNode)
                End If
            Next

            'If (candidates.Count = 0) Then Throw New Exception("candidates List has size 0 just before return in GetNodeToAdd")
            Return candidates(rnd.[Next](0, candidates.Count))
            ' if candidates has Count 1 we'll get that one node
        End Function ' GetNodeToAdd

        Private Function GetNodeToDrop(graph As MyGraph, clique As List(Of Integer), oneMissing As List(Of Integer)) As Integer
            ' get a node from clique set, which if dropped, gives the largest increase in PA set size
            ' we use the oneMissingb set to determine which clique node to pick
            'If (IsNothing(clique)) Then Throw New Exception("clique is null in GetNodeToDrop")
            'If (clique.Count = 0) Then Throw New Exception("clique has Count 0 in GetNodeToDrop")

            If clique.Count = 1 Then
                Return clique(0)
            End If

            ' scan each node in clique and determine the max possibleAdd size if node removed
            Dim maxCount As Integer = 0
            ' see explanation below
            For i As Integer = 0 To clique.Count - 1
                ' each node in clique nodes List
                Dim currCliqueNode As Integer = clique(i)
                Dim countNotAdjacent As Integer = 0
                For j As Integer = 0 To oneMissing.Count - 1
                    ' each node in the one missing list
                    Dim currOneMissingNode As Integer = oneMissing(j)
                    If graph.AreAdjacent(currCliqueNode, currOneMissingNode) = False Then
                        ' we like this
                        countNotAdjacent += 1

                        ' if currCliqueNode is not connected to omNode then currCliqueNode is the 'missing'
                        ' it would be good to drop this cliqueNode because after dropped from clique
                        ' the remaining nodes in the clique will all be connected to the omNode
                        ' and so the omNode would become a posibleAdd node and increase PA set size
                        ' So the best node to drop from clique will be the one which is least connected
                        ' to the nodes in OM
                    End If
                Next

                If countNotAdjacent > maxCount Then
                    maxCount = countNotAdjacent
                End If
            Next

            ' at this point we know what the max-not-connected count is but there could be several clique nodes which give that size
            Dim candidates As New List(Of Integer)()
            For i As Integer = 0 To clique.Count - 1
                ' each node in clique
                Dim currCliqueNode As Integer = clique(i)
                Dim countNotAdjacent As Integer = 0
                For j As Integer = 0 To oneMissing.Count - 1
                    ' each node in the one missing list
                    Dim currOneMissingNode As Integer = oneMissing(j)
                    If graph.AreAdjacent(currCliqueNode, currOneMissingNode) = False Then
                        countNotAdjacent += 1
                    End If
                Next

                If countNotAdjacent = maxCount Then
                    ' current clique node has max count not connected
                    candidates.Add(currCliqueNode)
                End If
            Next

            'If (candidates.Count = 0) Then Throw New Exception("candidates List has size 0 just before return in GetNodeToDropFromAllowedInClique")
            Return candidates(rnd.[Next](0, candidates.Count))
            ' must have size of at least 1
        End Function ' GetNodeToDrop
        
        Private Function MakeOneMissing(graph As MyGraph, clique As List(Of Integer)) As List(Of Integer)
            ' make a list of nodes in graph which are connected to all but one of the nodes in clique
            Dim count As Integer
            ' number of nodes in clique which are connected to a candidate node. if final count == (clique size - 1) then candidate is a winner
            Dim result As New List(Of Integer)()

            For i As Integer = 0 To graph.NumberNodes - 1
                ' each node in graph i a candidate
                count = 0
                If graph.NumberNeighbors(i) < clique.Count - 1 Then
                    Continue For
                End If
                ' node i has too few neighbors to possibly be connected to all but 1 node in clique
                'if (LinearSearch(clique, i) == true) continue; // node i is in clique. clique is not sorted so use LinearSearch -- consider Sort + BinarySearch
                If clique.BinarySearch(i) >= 0 Then
                    Continue For
                End If
                For j As Integer = 0 To clique.Count - 1
                    ' count how many nodes in clique are connected to candidate i
                    If graph.AreAdjacent(i, clique(j)) Then
                        count += 1
                    End If
                Next
                If count = clique.Count - 1 Then
                    result.Add(i)
                End If
            Next
            ' each candidate node i
            Return result

        End Function ' MakeOneMissing

        Private Function ListAsString(list As List(Of Integer)) As String
            Dim s As String = ""
            For i As Integer = 0 To list.Count - 1
                If i Mod 10 = 0 AndAlso i > 0 Then
                    s += Environment.NewLine
                End If
                s += list(i) & " "
            Next

            s += Environment.NewLine
            Return s
        End Function ' ListAsString

        Private Sub ValidateState(graph As MyGraph, clique As List(Of Integer), possibleAdd As List(Of Integer), oneMissing As List(Of Integer))
            ' any duplicates in clique?
            For i As Integer = 0 To clique.Count - 2
                For j As Integer = i + 1 To clique.Count - 1
                    If clique(i) = clique(j) Then
                        Throw New Exception("Dup value in clique")
                    End If
                Next
            Next
            ' any values in clique in possibleAdd or oneMissing?
            For i As Integer = 0 To possibleAdd.Count - 1
                If clique.BinarySearch(possibleAdd(i)) >= 0 Then
                    Throw New Exception("Possible Add value in clique")
                End If
            Next
            For i As Integer = 0 To oneMissing.Count - 1
                If clique.BinarySearch(oneMissing(i)) >= 0 Then
                    Throw New Exception("One Missing value in clique")
                End If
            Next
            ' any values in possibleAdd in oneMissing?
            For i As Integer = 0 To possibleAdd.Count - 1
                If oneMissing.Contains(possibleAdd(i)) Then
                    Throw New Exception("Value in possibleAdd found in oneMissing")
                End If
            Next
            For i As Integer = 0 To oneMissing.Count - 1
                If possibleAdd.Contains(oneMissing(i)) Then
                    Throw New Exception("Value in oneMissing found in possibleAdd")
                End If
            Next
        End Sub ' ValidateState
    End Module ' class Program

    Public Class MyGraph
        Private data As BitMatrix
        Private m_numberNodes As Integer
        Private m_numberEdges As Integer
        Private m_numberNeighbors As Integer()

        Public Sub New(graphFile As String, fileFormat As String)
            If fileFormat.ToUpper() = "DIMACS" Then
                LoadDimacsFormatGraph(graphFile)
            Else
                Throw New Exception("Format " + fileFormat + " not supported")
            End If
        End Sub

        Private Sub LoadDimacsFormatGraph(graphFile As String)
            Dim ifs As New FileStream(graphFile, FileMode.Open)
            Dim sr As New StreamReader(ifs)
            Dim line As String = ""
            Dim tokens As String() = Nothing

            ' advance to and get the p line (ex: "p edge 9 16")
            line = sr.ReadLine()
            ' read first line of file as a priming read
            line = line.Trim()
            While line IsNot Nothing AndAlso line.StartsWith("p") = False
                line = sr.ReadLine()
                line = line.Trim()
            End While

            tokens = line.Split(" "c)
            Dim numNodes As Integer = Integer.Parse(tokens(2))
            ' number nodes
            Dim numEdges As Integer = Integer.Parse(tokens(3))
            ' number edges
            sr.Close()
            ifs.Close()

            Me.data = New BitMatrix(numNodes)

            ifs = New FileStream(graphFile, FileMode.Open)
            ' reopen file
            sr = New StreamReader(ifs)
            While (InlineAssignHelper(line, sr.ReadLine())) IsNot Nothing
                line = line.Trim()
                If line.StartsWith("e") = True Then
                    ' (ex: "e 7 4")
                    tokens = line.Split(" "c)
                    Dim nodeA As Integer = Integer.Parse(tokens(1)) - 1
                    ' DIMACS is 1-based. subtract 1 to convert to 0-based
                    Dim nodeB As Integer = Integer.Parse(tokens(2)) - 1
                    data.SetValue(nodeA, nodeB, True)
                    data.SetValue(nodeB, nodeA, True)
                End If
            End While
            sr.Close()
            ifs.Close()

            Me.m_numberNeighbors = New Integer(numNodes - 1) {}
            For row As Integer = 0 To numNodes - 1
                Dim count As Integer = 0
                For col As Integer = 0 To numNodes - 1
                    If data.GetValue(row, col) = True Then
                        count += 1
                    End If
                Next
                m_numberNeighbors(row) = count
            Next

            Me.m_numberNodes = numNodes
            Me.m_numberEdges = numEdges
            Return
        End Sub

        Public ReadOnly Property NumberNodes() As Integer
            Get
                Return Me.m_numberNodes
            End Get
        End Property

        Public ReadOnly Property NumberEdges() As Integer
            Get
                Return Me.m_numberEdges
            End Get
        End Property

        Public Function NumberNeighbors(node As Integer) As Integer
            Return Me.m_numberNeighbors(node)
        End Function

        Public Function AreAdjacent(nodeA As Integer, nodeB As Integer) As Boolean
            If Me.data.GetValue(nodeA, nodeB) = True Then
                Return True
            Else
                Return False
            End If
        End Function

        Public Overrides Function ToString() As String
            Dim s As String = ""
            For i As Integer = 0 To Me.data.[Dim] - 1
                s += i + ": "
                For j As Integer = 0 To Me.data.[Dim] - 1
                    If Me.data.GetValue(i, j) = True Then
                        s += j + " "
                    End If
                Next
                s += Environment.NewLine
            Next
            Return s
        End Function

        Public Shared Sub ValidateGraphFile(graphFile As String, fileFormat As String)
            If fileFormat.ToUpper() = "DIMACS" Then
                ValidateDimacsGraphFile(graphFile)
            Else
                Throw New Exception("Format " + fileFormat + " not supported")
            End If
        End Sub

        Public Shared Sub ValidateDimacsGraphFile(graphFile As String)
            Dim ifs As New FileStream(graphFile, FileMode.Open)
            Dim sr As New StreamReader(ifs)
            Dim line As String = ""
            Dim tokens As String() = Nothing

            While (InlineAssignHelper(line, sr.ReadLine())) IsNot Nothing
                line = line.Trim()
                If line.StartsWith("c") = False AndAlso line.StartsWith("p") = False AndAlso line.StartsWith("e") = False Then
                    Throw New Exception("Unknown line type: " + line + " in ValidateDimacsGraphFile")
                End If

                Try
                    If line.StartsWith("p") Then
                        tokens = line.Split(" "c)
                        Dim numNodes As Integer = Integer.Parse(tokens(2))
                        Dim numEdges As Integer = Integer.Parse(tokens(3))
                    ElseIf line.StartsWith("e") Then
                        tokens = line.Split(" "c)
                        Dim nodeA As Integer = Integer.Parse(tokens(1))
                        Dim nodeB As Integer = Integer.Parse(tokens(2))
                    End If
                Catch
                    Throw New Exception("Error parsing line = " + line + " in ValidateDimacsGraphFile")
                End Try
            End While

            sr.Close()
            ifs.Close()
            Return
        End Sub ' ValidateDimacsGraphFile

        Public Sub ValidateGraph()
            ' total number edges items must be an even number
            Dim numConnections As Integer = 0
            For i As Integer = 0 To Me.data.[Dim] - 1
                For j As Integer = 0 To Me.data.[Dim] - 1
                    If Me.data.GetValue(i, j) = True Then
                        numConnections += 1
                    End If
                Next
            Next
            If numConnections Mod 2 <> 0 Then
                Throw New Exception("Total number of connections in graph is " + numConnections + ". Should be even")
            End If

            ' fully symmetric
            For i As Integer = 0 To Me.data.[Dim] - 1
                If Me.data.GetValue(i, i) = True Then
                    Throw New Exception("Node " + i + " is connected to itself")
                End If
                For j As Integer = 0 To Me.data.[Dim] - 1
                    If Me.data.GetValue(i, j) <> Me.data.GetValue(j, i) Then
                        Throw New Exception("Graph is not symmetric at " + i + " and " + j)
                    End If
                Next
            Next
            Return
        End Sub ' ValidateGraph

        Private Class BitMatrix
            Private data As BitArray()
            Public ReadOnly [Dim] As Integer

            Public Sub New(n As Integer)
                Me.data = New BitArray(n - 1) {}
                For i As Integer = 0 To data.Length - 1
                    Me.data(i) = New BitArray(n)
                Next
                Me.[Dim] = n
            End Sub

            Public Function GetValue(row As Integer, col As Integer) As Boolean
                Return data(row)(col)
            End Function

            Public Sub SetValue(row As Integer, col As Integer, value As Boolean)
                data(row)(col) = value
            End Sub

            Public Overrides Function ToString() As String
                Dim s As String = ""
                For i As Integer = 0 To data.Length - 1
                    For j As Integer = 0 To data(i).Length - 1
                        If data(i)(j) = True Then
                            s += "1 "
                        Else
                            s += "0 "
                        End If
                    Next
                    s += Environment.NewLine
                Next
                Return s
            End Function

        End Class ' class BitMatrix

        Private Shared Function InlineAssignHelper(Of T)(ByRef target As T, value As T) As T
            target = value
            Return value
        End Function
    End Class ' class MyGraph
End Namespace
